@@ -6,12 +6,13 @@
 
 __author__ = """luphord"""
 __email__ = """luphord@protonmail.com"""
-__version__ = """0.2.0"""
+__version__ = """0.3.0"""
 
 
 from argparse import ArgumentParser, Namespace
 from abc import ABC, abstractmethod
-from typing import Final, Union, Mapping, Tuple
+from typing import Final, Union, Mapping, Tuple, Set
+from numbers import Real
 import numpy as np
 from dataclasses import dataclass
 
@@ -33,7 +34,7 @@ class SimulatedCashflows:
             cashflows.dtype == self.dtype
         ), f"Got cashflow array with dtype {cashflows.dtype}, expecting {self.dtype}"
         assert cashflows.ndim == 2, f"Array must have ndim 2, got {cashflows.ndim}"
-        assert currencies.dtype == (np.string_, _ccy_letters)
+        assert currencies.dtype == (np.unicode_, _ccy_letters)
         assert currencies.shape == (cashflows.shape[1],)
         self.cashflows = cashflows
         self.currencies = currencies
@@ -56,7 +57,7 @@ class IndexedCashflows:
             cashflows.dtype == self.dtype
         ), f"Got cashflow array with dtype {cashflows.dtype}, expecting {self.dtype}"
         assert cashflows.ndim == 2, f"Array must have ndim 2, got {cashflows.ndim}"
-        assert currencies.dtype == (np.string_, _ccy_letters)
+        assert currencies.dtype == (np.unicode_, _ccy_letters)
         assert currencies.shape == (cashflows.shape[1],)
         self.cashflows = cashflows
         self.currencies = currencies
@@ -86,6 +87,14 @@ class IndexedCashflows:
     def __neg__(self) -> "IndexedCashflows":
         return self * -1
 
+    def zero_after(self, date_idx: "DateIndex") -> "IndexedCashflows":
+        assert self.nsim == date_idx.nsim
+        zeroedcf = self.cashflows.copy()
+        for i, cf in enumerate(self.cashflows.T):
+            ko_mask = (cf["index"] >= date_idx.index) & (date_idx.index >= 0)
+            zeroedcf["value"][ko_mask, i] = 0
+        return IndexedCashflows(zeroedcf, self.currencies, self.dategrid)
+
     def apply_index(self) -> SimulatedCashflows:
         dategrid_rep = np.reshape(self.dategrid, (1, self.dategrid.size))
         dategrid_rep = np.repeat(dategrid_rep, self.nsim, axis=0)
@@ -163,6 +172,14 @@ class Model:
         self.shape = (self.nsim, self.ndates)
         assert numeraire.shape == self.shape
         self.numeraire_currency = numeraire_currency
+        for fxkey, val in simulated_fx.items():
+            assert (
+                val.dtype == np.float
+            ), f"FX spot '{fxkey}' is of dtype {val.dtype}, expecting float"
+            assert (
+                val.shape == self.shape
+            ), f"FX Spot '{fxkey}' has shape {val.shape}, expecting {self.shape}"
+        self.simulated_fx = simulated_fx
         for key, val in simulated_stocks.items():
             assert (
                 val.dtype == np.float
@@ -170,7 +187,6 @@ class Model:
             assert (
                 val.shape == self.shape
             ), f"Stock '{key}' has shape {val.shape}, expecting {self.shape}"
-        self.simulated_fx = simulated_fx
         self.simulated_stocks = simulated_stocks
 
     @property
@@ -181,7 +197,22 @@ class Model:
     def eval_date_index(self) -> DateIndex:
         return DateIndex(np.zeros((self.nsim,), dtype=np.int))
 
+    @property
+    def currencies(self) -> Set[str]:
+        return set([ccy for fxspot in self.simulated_fx for ccy in fxspot]) or {
+            self.numeraire_currency
+        }
+
     def get_simulated_fx(self, base_currency: str, counter_currency: str) -> np.ndarray:
+        assert (
+            base_currency != _null_ccy and counter_currency != _null_ccy
+        ), "Cannot retrieve spot containing null currency NNN"
+        assert (
+            base_currency in self.currencies
+        ), f"{base_currency} not contained in model currencies {self.currencies}"
+        assert (
+            counter_currency in self.currencies
+        ), f"{counter_currency} not contained in model currencies {self.currencies}"
         if (base_currency, counter_currency) in self.simulated_fx:
             return self.simulated_fx[(base_currency, counter_currency)]
         elif (counter_currency, base_currency) in self.simulated_fx:
@@ -196,15 +227,68 @@ class Model:
     def generate_cashflows(self, contract: "Contract") -> SimulatedCashflows:
         return contract.generate_cashflows(self.eval_date_index, self).apply_index()
 
+    def in_currency(
+        self, cashflows: IndexedCashflows, currency: str
+    ) -> IndexedCashflows:
+        assert currency != _null_ccy, "Cannot convert to null currency NNN"
+        currencies = np.zeros(
+            cashflows.currencies.shape, dtype=(np.unicode_, _ccy_letters)
+        )
+        currencies[:] = currency
+        converted = np.zeros(cashflows.cashflows.shape, dtype=IndexedCashflows.dtype)
+        for i, cf in enumerate(cashflows.cashflows.T):
+            converted["index"][:, i] = cf["index"]
+            if cashflows.currencies[i] == _null_ccy:
+                currencies[i] = _null_ccy
+            if cashflows.currencies[i] == currencies[i]:
+                converted["value"][:, i] = cf["value"]
+            else:
+                base_ccy = str(cashflows.currencies[i])
+                counter_ccy = str(currencies[i])
+                fx = self.get_simulated_fx(base_ccy, counter_ccy)
+                fx_indexed = DateIndex(cf["index"]).index_column(fx)
+                converted["value"][:, i] = cf["value"] * fx_indexed
+        return IndexedCashflows(converted, currencies, cashflows.dategrid)
+
+    def in_numeraire_currency(self, cashflows: IndexedCashflows) -> IndexedCashflows:
+        return self.in_currency(cashflows, self.numeraire_currency)
+
 
 class ObservableFloat(ABC):
+    """Abstract base class for all observables of underlying type float,
+    essentially a real-valued stochastic process"""
+
     @abstractmethod
     def simulate(self, model: Model) -> np.ndarray:
         pass
 
+    def __ge__(self, other: Union["ObservableFloat", float]) -> "ObservableBool":
+        if isinstance(other, ObservableFloat):
+            return GreaterOrEqualThan(self, other)
+        elif isinstance(other, Real):
+            return GreaterOrEqualThan(self, KonstFloat(other))
+        else:
+            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+
+    def __gt__(self, other: Union["ObservableFloat", float]) -> "ObservableBool":
+        if isinstance(other, ObservableFloat):
+            return GreaterThan(self, other)
+        elif isinstance(other, Real):
+            return GreaterThan(self, KonstFloat(other))
+        else:
+            raise TypeError(f"Expecting real number, got {other} of type {type(other)}")
+
+    def __le__(self, other: Union["ObservableFloat", float]) -> "ObservableBool":
+        return ~(self > other)
+
+    def __lt__(self, other: Union["ObservableFloat", float]) -> "ObservableBool":
+        return ~(self >= other)
+
 
 @dataclass
 class Stock(ObservableFloat):
+    """Value of the stock identified by identifier"""
+
     identifier: str
 
     def simulate(self, model: Model) -> np.ndarray:
@@ -213,6 +297,9 @@ class Stock(ObservableFloat):
 
 @dataclass
 class FX(ObservableFloat):
+    """Value of the currency spot between base_currency and counter_currency,
+    i.e. 'one unit counter_currency' / 'one unit of base_currency'"""
+
     base_currency: str
     counter_currency: str
 
@@ -222,6 +309,8 @@ class FX(ObservableFloat):
 
 @dataclass
 class KonstFloat(ObservableFloat):
+    """Always equal to constant"""
+
     constant: float
 
     def simulate(self, model: Model) -> np.ndarray:
@@ -229,13 +318,80 @@ class KonstFloat(ObservableFloat):
 
 
 class ObservableBool(ABC):
+    """Abstract base class for all observables of underlying type bool"""
+
     @abstractmethod
     def simulate(self, model: Model) -> np.ndarray:
         pass
 
+    def __invert__(self) -> "ObservableBool":
+        return Not(self)
+
+    def __and__(self, other: "ObservableBool") -> "ObservableBool":
+        return AndObservable(self, other)
+
+    def __or__(self, other: "ObservableBool") -> "ObservableBool":
+        return OrObservable(self, other)
+
+
+@dataclass
+class Not(ObservableBool):
+    """True if observable is False and vice versa"""
+
+    observable: ObservableBool
+
+    def simulate(self, model: Model) -> np.ndarray:
+        return ~self.observable.simulate(model)
+
+
+@dataclass
+class AndObservable(ObservableBool):
+    """True if and only if both observables are True"""
+
+    observable1: ObservableBool
+    observable2: ObservableBool
+
+    def simulate(self, model: Model) -> np.ndarray:
+        return self.observable1.simulate(model) & self.observable2.simulate(model)
+
+
+@dataclass
+class OrObservable(ObservableBool):
+    """True if either or both observable are True"""
+
+    observable1: ObservableBool
+    observable2: ObservableBool
+
+    def simulate(self, model: Model) -> np.ndarray:
+        return self.observable1.simulate(model) | self.observable2.simulate(model)
+
+
+@dataclass
+class GreaterOrEqualThan(ObservableBool):
+    """True if and only if observable1 is greater or equal than observable2"""
+
+    observable1: ObservableFloat
+    observable2: ObservableFloat
+
+    def simulate(self, model: Model) -> np.ndarray:
+        return self.observable1.simulate(model) >= self.observable2.simulate(model)
+
+
+@dataclass
+class GreaterThan(ObservableBool):
+    """True if and only if observable1 is strictly greater than observable2"""
+
+    observable1: ObservableFloat
+    observable2: ObservableFloat
+
+    def simulate(self, model: Model) -> np.ndarray:
+        return self.observable1.simulate(model) > self.observable2.simulate(model)
+
 
 @dataclass
 class At(ObservableBool):
+    """True only at date"""
+
     date: np.datetime64
 
     def simulate(self, model: Model) -> np.ndarray:
@@ -245,6 +401,8 @@ class At(ObservableBool):
 
 
 class Contract(ABC):
+    """Abstract base class for all contracts"""
+
     @abstractmethod
     def generate_cashflows(
         self, acquisition_idx: DateIndex, model: Model
@@ -254,18 +412,22 @@ class Contract(ABC):
 
 @dataclass
 class Zero(Contract):
+    """Neither receive nor pay anything"""
+
     def generate_cashflows(
         self, acquisition_idx: DateIndex, model: Model
     ) -> IndexedCashflows:
         model.validate_date_index(acquisition_idx)
         cf = np.zeros((model.nsim, 1), dtype=IndexedCashflows.dtype)
         cf["index"][:, 0] = acquisition_idx.index
-        ccys = np.array([_null_ccy], dtype=(np.string_, _ccy_letters))
+        ccys = np.array([_null_ccy], dtype=(np.unicode_, _ccy_letters))
         return IndexedCashflows(cf, ccys, model.dategrid)
 
 
 @dataclass
 class One(Contract):
+    """Receive one unit of currency at acquisition"""
+
     currency: str
 
     def generate_cashflows(
@@ -276,12 +438,15 @@ class One(Contract):
         cf["index"][:, 0] = acquisition_idx.index
         cf["index"][acquisition_idx.index < 0, 0] = -1  # index < 0 means never
         cf["value"][acquisition_idx.index < 0, 0] = 0  # index < 0 means never
-        ccys = np.array([self.currency], dtype=(np.string_, _ccy_letters))
+        ccys = np.array([self.currency], dtype=(np.unicode_, _ccy_letters))
         return IndexedCashflows(cf, ccys, model.dategrid)
 
 
 @dataclass
 class Give(Contract):
+    """Receive all obligations of the underlying contract and pay all rights,
+    i.e. invert the underlying contract"""
+
     contract: Contract
 
     def generate_cashflows(
@@ -292,6 +457,8 @@ class Give(Contract):
 
 @dataclass
 class And(Contract):
+    """Obtain rights and obligations of both underlying contracts"""
+
     contract1: Contract
     contract2: Contract
 
@@ -305,6 +472,8 @@ class And(Contract):
 
 @dataclass
 class Or(Contract):
+    """Choose at acquisition between the underlying contracts"""
+
     contract1: Contract
     contract2: Contract
 
@@ -316,31 +485,27 @@ class Or(Contract):
         ccys = np.unique(
             np.concatenate((cf1.currencies.flatten(), cf2.currencies.flatten()))
         )
-        ccys = ccys[ccys != b"NNN"].flatten()
+        ccys = ccys[ccys != "NNN"].flatten()
         for cf in (cf1 + cf2).cashflows.T:
             if (acquisition_idx.index != cf["index"]).any():
                 raise NotImplementedError(
                     "Cashflow generation for OR contract at any moment"
                     " other than cashflow date is not implemented"
                 )
-        if ccys.size > 1:
-            raise NotImplementedError(
-                "Cashflow generation for OR contract for different currencies"
-                f" is not implemented, got {ccys}"
-            )
-        else:
-            cf1sum = cf1.cashflows["value"].sum(axis=1)
-            cf2sum = cf2.cashflows["value"].sum(axis=1)
-            choose1 = cf1sum > cf2sum
-            cf1.cashflows["index"][~choose1] = -1
-            cf1.cashflows["value"][~choose1] = 0
-            cf2.cashflows["index"][choose1] = -1
-            cf2.cashflows["value"][choose1] = 0
-            return cf1 + cf2
+        cf1sum = model.in_numeraire_currency(cf1).cashflows["value"].sum(axis=1)
+        cf2sum = model.in_numeraire_currency(cf2).cashflows["value"].sum(axis=1)
+        choose1 = cf1sum > cf2sum
+        cf1.cashflows["index"][~choose1] = -1
+        cf1.cashflows["value"][~choose1] = 0
+        cf2.cashflows["index"][choose1] = -1
+        cf2.cashflows["value"][choose1] = 0
+        return cf1 + cf2
 
 
 @dataclass
 class Cond(Contract):
+    """If observable is True at acquisition, obtain contract1, otherwise contract2"""
+
     observable: ObservableBool
     contract1: Contract
     contract2: Contract
@@ -363,6 +528,9 @@ class Cond(Contract):
 
 @dataclass
 class Scale(Contract):
+    """Same as the underling contract, but all payments scaled by the value of
+    observable at acquisition"""
+
     observable: ObservableFloat
     contract: Contract
 
@@ -380,6 +548,9 @@ class Scale(Contract):
 
 @dataclass
 class When(Contract):
+    """Obtain the underlying contract as soon as observable
+    becomes True after acquisition"""
+
     observable: ObservableBool
     contract: Contract
 
@@ -392,6 +563,9 @@ class When(Contract):
 
 @dataclass
 class Anytime(Contract):
+    """At any point in time after acquisition when observable is True, choose whether
+    to obtain the underlying contract or not; can be exercised only once"""
+
     observable: ObservableBool
     contract: Contract
 
@@ -403,13 +577,18 @@ class Anytime(Contract):
 
 @dataclass
 class Until(Contract):
+    """Obtain the underlying contract, but as soon as observable
+    becomes True after acquisition all following payments are nullified"""
+
     observable: ObservableBool
     contract: Contract
 
     def generate_cashflows(
         self, acquisition_idx: DateIndex, model: Model
     ) -> IndexedCashflows:
-        raise NotImplementedError()
+        cf = self.contract.generate_cashflows(acquisition_idx, model)
+        ko_idx = acquisition_idx.next_after(self.observable.simulate(model))
+        return cf.zero_after(ko_idx)
 
 
 class ResolvableContract(Contract):
@@ -444,6 +623,42 @@ class EuropeanOption(ResolvableContract):
         return When(At(self.maturity), Or(self.contract, Zero()))
 
 
+class BrownianMotion:
+    """Brownian Motion (Wiener Process) with optional drift."""
+
+    def __init__(self, mu: float = 0.0, sigma: float = 1.0):
+        self.mu = mu
+        self.sigma = sigma
+
+    def simulate(self, t: np.array, n: int, rnd: np.random.RandomState) -> np.array:
+        assert t.ndim == 1, "One dimensional time vector required"
+        assert t.size > 0, "At least one time point is required"
+        dt = np.concatenate((t[0:1], np.diff(t)))
+        assert (dt >= 0).all(), "Increasing time vector required"
+        # transposed simulation for automatic broadcasting
+        W = rnd.normal(size=(n, t.size))
+        W_drift = W * np.sqrt(dt) * self.sigma + self.mu * dt
+        return np.cumsum(W_drift, axis=1)
+
+
+class GeometricBrownianMotion:
+    """Geometric Brownian Motion.(with optional drift)."""
+
+    def __init__(self, mu: float = 0.0, sigma: float = 1.0):
+        self.mu = mu
+        self.sigma = sigma
+
+    def simulate(self, t: np.array, n: int, rnd: np.random.RandomState) -> np.array:
+        assert t.ndim == 1, "One dimensional time vector required"
+        assert t.size > 0, "At least one time point is required"
+        dt = np.concatenate((t[0:1], np.diff(t)))
+        assert (dt >= 0).all(), "Increasing time vector required"
+        # transposed simulation for automatic broadcasting
+        dW = (rnd.normal(size=(t.size, n)).T * np.sqrt(dt)).T
+        W = np.cumsum(dW, axis=0)
+        return np.exp(self.sigma * W.T + (self.mu - self.sigma ** 2 / 2) * t)
+
+
 parser = ArgumentParser(description=__doc__)
 parser.add_argument(
     "--version", help="Print version number", default=False, action="store_true"
